from typing import Optional
import boto3
import time
from shared_kernel.batch_job.aws_batch.job_status_mapper import JOB_STATUS_TO_TASK_STATUS
from shared_kernel.batch_job.s3.s3_file_handler import S3FileHandler
from shared_kernel.config import Config
from shared_kernel.dataclasses.batch_job_config import BatchJobConfig
from shared_kernel.exceptions.batch_job_exceptions import BatchJobFailedException
from shared_kernel.logger import Logger
from shared_kernel.status_tracker.status_tracker import StatusTracker
from shared_kernel.enums.status_tracker import TaskStatus


config = Config()
logger = Logger(config.get("APP_NAME"))
AWS_REGION =  config.get("AWS_REGION")
DATALAKE_BUCKET = config.get("FDC_DATA_LAKE_BUCKET")
BATCH_JOBS_PREFIX = "batch_jobs"

class AWSBatchJobHandler:
    """
    Handles submission, tracking, and result retrieval of AWS Batch jobs using S3 for payload exchange.

    Core Responsibilities:
    - Submit AWS Batch jobs with required configuration
    - Poll and wait for job status until completion/failure
    - Upload input payload to S3 and fetch result from S3
    - Track and update task status in StatusTracker
    """
    def __init__(self):
        self.batch_client = boto3.client("batch", region_name=AWS_REGION)
        self.s3_file_handler = S3FileHandler()
        self.status_tracker = StatusTracker()

    def submit_job(
        self,
        job_config: BatchJobConfig
    ) -> str:
        """
        Submit a job to AWS Batch with the provided configuration.

        Args:
            job_config (BatchJobConfig): Configuration object containing job name, queue, definition, and environment variables.

        Returns:
            str: The AWS Batch job ID.

        Raises:
            BatchJobFailedException: If job submission fails or returns no job ID.
            Exception: For unexpected failures.
        """
        try:
            env_vars = [{"name": k, "value": v} for k, v in job_config.to_env_vars().items()]
            overrides = job_config.container_overrides or {}
            overrides["environment"] = env_vars
            job_payload = {
                "jobName": job_config.job_name,
                "jobQueue": job_config.job_queue,
                "jobDefinition": job_config.job_definition,
                "containerOverrides": overrides,
            }
            response = self.batch_client.submit_job(**job_payload)
            job_id = response["jobId"]
            if not job_id:
                raise BatchJobFailedException(
                    job_type=job_config.job_type,
                    job_id=None,
                    job_name=job_config.job_name,
                    reason="Batch job submission did not return a job ID."
                )
            job_config.batch_job_id = job_id
            logger.info(f"Submitted AWS Batch job {job_config.job_name} with ID {job_id}")
            self.status_tracker.create_task(event_msg=job_config.to_status_event(), status=TaskStatus.QUEUED.value)
            return job_id
        except BatchJobFailedException as e:
            raise
        except Exception as e:
            logger.error(f"Failed to submit AWS Batch job {job_config.job_name}: {str(e)}")
            raise

    def wait_for_job_completion(
        self,
        job_config: BatchJobConfig,
        timeout: int = 180,
        poll_interval: int = 10
    ) -> None:
        """
        Polls AWS Batch to wait for the completion of a submitted job.

        Args:
            job_config (BatchJobConfig): The job configuration with batch job ID.
            timeout (int): Maximum time (in seconds) to wait for job completion. Default is 180.
            poll_interval (int): Time (in seconds) between polling attempts. Default is 10.

        Raises:
            TimeoutError: If the job does not complete within the timeout.
            BatchJobFailedException: If the job fails during execution.
            Exception: For unexpected errors during polling.
        """
        start_time = time.time()
        while True:
            try:
                response = self.batch_client.describe_jobs(jobs=[job_config.batch_job_id])
                job_detail = response["jobs"][0]
                status_str = job_detail["status"]
                new_status = JOB_STATUS_TO_TASK_STATUS.get(status_str)
                if new_status and new_status != job_config.job_status:
                    logger.info(f"Job {job_config.batch_job_id} status changed from {job_config.job_status} to {new_status.value}")
                    self.status_tracker.update_task(
                        event_msg=job_config.to_status_event(),
                        status=new_status.value
                    )
                    job_config.job_status = new_status

                if status_str == "SUCCEEDED":
                    logger.info(f"Job {job_config.batch_job_id} completed successfully.")
                    return

                if status_str == "FAILED":
                    self.handle_failed_batch_job(job_config=job_config, job_detail=job_detail)

                if time.time() - start_time > timeout:
                    logger.error(f"Timeout waiting for job {job_config.batch_job_id}")
                    raise TimeoutError(f"Job {job_config.batch_job_id} did not complete within {timeout} seconds.")

                time.sleep(poll_interval)
            except BatchJobFailedException as e:
                raise
            except Exception as e:
                logger.error(f"Error while polling job {job_config.batch_job_id}: {str(e)}")
                raise

    def get_s3_path(self, job_config: BatchJobConfig) -> str:
        """
        Constructs the S3 path based on job type and folder key.

        Args:
            job_config (BatchJobConfig): The job configuration containing type and S3 folder key.

        Returns:
            str: Full S3 path prefix for the job.
        """
        return f"{BATCH_JOBS_PREFIX}/{job_config.job_type}/{job_config.s3_folder_key}"

    def _upload_payload_to_s3(self, payload: dict, job_config: BatchJobConfig) -> None:
        """
        Uploads the job input payload to S3.

        Args:
            payload (dict): The job input payload.
            job_config (BatchJobConfig): The job configuration including path info.

        Raises:
            Exception: If the S3 upload fails.
        """
        s3_key = self.get_s3_path(job_config=job_config)
        s3_input_key = f"{s3_key}/input.json"
        if not self.s3_file_handler.upload_json(key=s3_input_key, data=payload):
            raise Exception(f"Failed to upload batch job input to s3")

        logger.info(f"Uploaded job input to S3 at key: {s3_input_key}")
    
    def _download_response_from_s3(self, job_config: BatchJobConfig) -> Optional[dict]:
        """
        Downloads the job result JSON from S3.

        Args:
            job_config (BatchJobConfig): Job configuration used to compute S3 path.

        Returns:
            Optional[dict]: The downloaded job result as a dictionary, or None if not found.
        """
        s3_key = self.get_s3_path(job_config=job_config)
        result = self.s3_file_handler.download_json(key=f"{s3_key}/output.json")
        return result
    
    def process_batch_job_with_s3(self, job_config: BatchJobConfig, payload: dict) -> dict:
        """
        Handles the full lifecycle of a batch job:
        - Uploads input to S3
        - Submits the job
        - Waits for completion
        - Downloads and returns the result

        Args:
            job_config (BatchJobConfig): The job configuration.
            payload (dict): Input payload to upload to S3.

        Returns:
            dict: The job's output data.

        Raises:
            BatchJobFailedException: If the job fails or no result is available.
        """
        try:
            logger.info(f"Starting batch job process for: {job_config.job_name}")
            self._upload_payload_to_s3(payload=payload, job_config=job_config)
            self.submit_job(job_config=job_config)
            self.wait_for_job_completion(job_config=job_config)
            result = self._download_response_from_s3(job_config=job_config)
            if not result or not isinstance(result, dict):
                raise BatchJobFailedException(
                    job_type=job_config.job_type,
                    job_id=job_config.batch_job_id,
                    job_name=job_config.job_name,
                    reason="Job succeeded but no output found in S3"
                )
            if result.get("error"):
                raise BatchJobFailedException(
                    job_type=job_config.job_type,
                    job_id=job_config.batch_job_id,
                    job_name=job_config.job_name,
                    reason=result.get("message")
                )
            logger.info(f"Completed batch job process for: {job_config.job_name}")
            return result.get("data")
        except BatchJobFailedException:
            raise
        except Exception as e:
            logger.error(f"Failed to fetch result for {job_config.job_name}: {str(e)}")
            raise BatchJobFailedException(
                job_type=job_config.job_type,
                job_id=job_config.batch_job_id,
                job_name=job_config.job_name,
                reason=f"{str(e)}"
            )
    
    def handle_failed_batch_job(self, job_detail: dict, job_config: BatchJobConfig):
        """
        Handles job failure by extracting the failure reason, logging the error, 
        updating the status tracker, and raising a BatchJobFailedException.

        Args:
            job_detail (dict): AWS Batch job detail dictionary.
            job_config (BatchJobConfig): The associated job configuration.

        Raises:
            BatchJobFailedException: Always raised after failure handling.
        """
        failure_reason = self._extract_failure_reason(job_detail=job_detail, job_config=job_config)
        logger.error(f"Job {job_config.batch_job_id} failed during execution due to {failure_reason}.")
        self.status_tracker.mark_task_as_failure(
            **job_config.get_failure_event_info(failure_reason)
        )
        raise BatchJobFailedException(job_type=job_config.job_type,
                                    job_id=job_config.batch_job_id,
                                    reason=failure_reason,
                                    job_name=job_config.job_name)
        
        
    def _extract_failure_reason(self, job_detail: dict, job_config: BatchJobConfig) -> str:
        """
        Extracts the reason a job failed using S3 result or AWS Batch status.

        Args:
            job_detail (dict): AWS Batch job detail.
            job_config (BatchJobConfig): Job configuration for S3 path access.

        Returns:
            str: The failure reason.
        """
        failure_reason = None

        # Step 1: Try downloading result from S3 (best-effort)
        try:
            result = self._download_response_from_s3(job_config=job_config)
            if result and result.get("error"):
                failure_reason = result.get("message")
        except Exception as e:
            logger.warning(f"[Failure Reason Fallback] Could not download failure result from S3 for job {job_config.batch_job_id}: {str(e)}")

        if not failure_reason:
            failure_reason = job_detail.get("statusReason")
        return failure_reason