from dataclasses import dataclass
from typing import Optional, Dict
from shared_kernel.batch_job.aws_batch.job_status_mapper import JOB_STATUS_TO_TASK_STATUS
from shared_kernel.messaging.utils.event_messages import EventMessage
from shared_kernel.enums.status_tracker import TaskStatus
from dataclasses import field


@dataclass
class BatchJobConfig:
    """
    Configuration container for submitting and tracking AWS Batch jobs.

    Attributes:
        task_id (str): Unique identifier for the task/job used for status tracking.
        job_name (str): Name used to register the job in AWS Batch.
        job_queue (str): The Batch job queue to which the job will be submitted.
        job_definition (str): AWS Batch job definition.
        s3_folder_key (str): S3 prefix where input/output for the job is stored.
        job_type (str): Type/category of the job.
        event_meta (Dict[str, str]): Metadata used in event messages and tracing.
        batch_job_id (Optional[str]): AWS Batch job ID once the job is submitted.
        metadata (Optional[Dict[str, str]]): Additional environment variables.
        container_overrides (Optional[Dict]): Overrides for vCPU/memory for the container.
        job_status (Optional[TaskStatus]): Internal representation of the job’s status.
    """
    task_id: str
    job_name: str
    job_queue: str
    job_definition: str
    s3_folder_key: str
    job_type: str
    event_meta: Dict[str, str] = field(default_factory=dict)
    batch_job_id: Optional[str] = None
    metadata: Optional[Dict[str, str]] = None  # Additional env vars like METRIC_ID.
    container_overrides: Optional[Dict] = None  # Optional vCPUs, memory overrides
    job_status: Optional[TaskStatus] = None

    def to_env_vars(self) -> Dict[str, str]:
        """
        Constructs environment variables to be passed into the container.

        Returns:
            Dict[str, str]: Environment variables including job type, S3 folder key, 
                            and optionally any metadata.
        """
        env = {
            "S3_FOLDER_KEY": self.s3_folder_key,
            "JOB_TYPE": self.job_type
        }
        if self.metadata:
            env.update({k.upper(): str(v) for k, v in self.metadata.items()})
        if self.job_type:
            env.update({"JOB_TYPE": self.job_type})
        return env
        
    def to_status_event(self):
        """
        Creates an EventMessage representing this job's status event for status tracking.

        Returns:
            EventMessage: The event message with job type and task metadata.
        """
        event = {
            "event_name": self.job_type,
            "event_meta": {
                **self.event_meta,
                "job_id": self.task_id
            }
        }
        event_message = EventMessage(raw_message=event)
        return event_message
        
    def update_job_status(self, status: str):
        """
        Maps AWS Batch job status to internal TaskStatus and updates the instance.

        Args:
            status (str): The AWS job status string (e.g., 'SUCCEEDED', 'FAILED').

        Raises:
            ValueError: If the AWS job status is not recognized.
        """
        mapped_status = JOB_STATUS_TO_TASK_STATUS.get(status)
        if not mapped_status:
            raise ValueError(f"Unrecognized AWS job status: {status}")
        self.job_status = mapped_status

    def get_failure_event_info(self, failure_reason: str) -> dict:
        """
        Constructs failure information to be saved in status tracker.

        Args:
            failure_reason (str): Description for the job failure.

        Returns:
            dict: Dictionary containing trace/span/task identifiers and failure reason.
        """
        return {
            "span_id": self.event_meta.get("span_id"),
            "trace_id": self.event_meta.get("trace_id"),
            "task": self.job_type,
            "failure_reason": failure_reason,
            "task_id": self.task_id
        }